/*
Connexion et discussion avec le serveur FTP
*/

#include <stdio.h>
#include <windows.h>
#include <winsock.h>
#include <stdarg.h>
#include <time.h>

#include "ftpsync.h"
#include "ftpclient.h"

int sock;

int total;
char sockbuf[256+1];

extern int debug;
FILE *logfile;

/*
*/
int ftp_init(struct projet *proj)
{
	if (ftp_connect(proj)==-1)
		return -1;

	if (ftp_login(proj)==-1)
	{
		ftp_disconnect();
		return -1;
	}

	return 0;
}

/*
Connexion au serveur FTP
*/
int ftp_connect(struct projet *proj)
{
	struct hostent *h;
	struct sockaddr_in serv;
	/*char buffer[128];*/

	if ((sock=socket(AF_INET, SOCK_STREAM, 0))==-1)
	{
		printf("Creation socket impossible\n");
		return -1;
	}

	if ((h=gethostbyname(proj->host))==NULL)
	{
		printf("Resolution de %s impossible\n", proj->host);
		return -1;
	}

	memset(&serv, 0, sizeof(struct sockaddr_in));

	serv.sin_family=AF_INET;
	serv.sin_port=htons(21);
	serv.sin_addr.s_addr=*((unsigned long *)h->h_addr_list[0]);

	printf("Connecting to %s...\n", proj->host);
	if (connect(sock, (struct sockaddr *)&serv, sizeof(struct sockaddr))==-1)
	{
		printf("La connexion a echoue\n");
		return -1;
	}

	if ((logfile=fopen("ftpsync.log", "a+"))==NULL)
		printf("Can't open logfile\n");

	return 0;
}

/*
Login / Pass
Changement du repertoire courant
*/
int ftp_login(struct projet *proj)
{
	total=0;
	*sockbuf=0x00;

	if (read_answer_code()!=220)
		return -1;
	
	skprintf("USER %s\r\n", proj->user);
	if (read_answer_code()!=331)
	{
		printf("Login: utilisateur refuse\n");
		return -1;
	}
	
	skprintf("PASS %s\r\n", proj->pass);
	if (read_answer_code()!=230)
	{
		printf("Login: mot de passe incorrect\n");
		return -1;
	}

	printf("Logged in !\n");

	if (*proj->host_dir!=0x00)
	{
		skprintf("CWD %s\r\n", proj->host_dir);
		if (read_answer_code()!=250)
			printf("CWD: erreur\n");
	}

	return 0;
}

/*
Retourne le code de reponse du serveur
*/
int read_answer_code(void)
{
	int code;
	char buffer[128];

	do
	{
		if (read_line(buffer)==-1)
			return -1;
		code=get_answer_code(buffer);
	}
	while (code==0);

	return code;
}

/*
Convertit la reponse du serveur
en un entier sign
*/
int get_answer_code(char *buffer)
{
	int i;
	int val;

	val=0;
	for(i=0; i<3; i++)
	{
		if (buffer[i]==0x00)
			return 0;
		if (buffer[i]<'0' || buffer[i]>'9')
			return 0;
		val=val*10+buffer[i]-'0';
	}
	if (buffer[3]==' ')
		return val;

	return 0;
}

/*
*/
int read_line(char *buffer)
{
	int r;
	char *p;

	while ((p=strchr(sockbuf, '\n'))==NULL)
	{
		if ((r=recv(sock, sockbuf+total, 256-total, 0))<=0)
		{
			printf("Deconnexion distante\n");
			return -1;
		}
		total+=r;
		sockbuf[total]=0x00;
	}
	
	*p=0x00;
	if (*(p-1)=='\r')
		*(p-1)=0x00;
	p++;
	strcpy(buffer, sockbuf);

	strcpy(sockbuf, p);
	total=strlen(sockbuf);

	if (debug)
		printf("%s\n", buffer);

	if (logfile)
		fprintf(logfile, "%s\n", buffer);
	
	return 0;
}

/*
Ecriture formatte sur une socket
*/
int skprintf(char *format, ...)
{
	int n;
	char buffer[256];
	va_list v;

	va_start(v, format);

	n=vsprintf(buffer, format, v);

	return send(sock, buffer, n, 0);
}

/*
Change de rpertoire
Cre le rpertoire s'il n'existe pas
*/
int chdir(char *dir)
{
	if (debug)
		printf("chdir %s\n", dir);

	skprintf("CWD %s\r\n", dir);
	if (read_answer_code()!=250)
	{
		printf("Creating directory %s...", dir);
		skprintf("MKD %s\r\n", dir);
		if (read_answer_code()!=257)
		{
			printf("failed\n");
			return -1;
		}
		skprintf("CWD %s\r\n", dir);
		if (read_answer_code()!=250)
			printf("failed\n");

		printf("ok\n");
	}
	return 0;
}

/*
Envoie un fichier
*/
int send_file(HANDLE hfile, struct file_properties *fprop)
{
	int datasock;
	int i;
	char buffer[512+1];
	struct sockaddr_in serv;
	DWORD b_read, total = 0;
	unsigned long ip;
	unsigned long tdeb;

	printf("Sending %s... ", fprop->local);

	/* TYPE IMAGE */
	skprintf("TYPE I\r\n");
	if (read_answer_code()!=200)
	{
		printf("Le serveur refuse de passer en mode image\n");
		return -1;
	}

	/* Mode passif: le serveur FTP met un port sur coute
	et le client s'y connecte pour envoyer les donnes */
	skprintf("PASV\r\n");
	do
	{
		if (read_line(buffer)==-1)
			return -1;
		i=get_answer_code(buffer);
	}
	while (i==0);

	if (i!=227)
	{
		printf("Mode passif non supporte !\n");
		return -1;
	}

	memset(&serv, 0, sizeof(struct sockaddr_in));

	serv.sin_family=AF_INET;
	serv.sin_port=htons(parse_pasv_cmd(buffer, &ip));
	serv.sin_addr.s_addr=ip;

	if (debug)
		printf("%s:%d\n", inet_ntoa(serv.sin_addr), htons(serv.sin_port));

	if ((datasock=socket(AF_INET, SOCK_STREAM, 0))==-1)
	{
		printf("Creation socket impossible\n");
		return -1;
	}

	skprintf("STOR %s\r\n", fprop->remote);
	
	if (debug)
		printf("STOR %s, attente\n", fprop->remote);

	i=3;
	while (connect(datasock, (struct sockaddr *)&serv, sizeof(struct sockaddr))==-1)
	{
		if (--i==0)
		{
			printf("failed\n");
			return -1;
		}
		printf(".");
	}

	if (read_answer_code()!=150)
	{
		printf("send_file() : reponse du serveur incorrecte\n");
		return -1;
	}

	/* -1 pour viter une division par zro */
	tdeb=time(NULL)-1;

	while (ReadFile(hfile, buffer, 512, &b_read, NULL))
	{
		if (b_read==0)
			break;
		if (send(datasock, buffer, b_read, 0)<=0)
		{
			printf("send_file: connexion perdue\n");
			/* return -1 */
			break;
		}
		total+=b_read;
	}

	closesocket(datasock);

	if (read_answer_code()!=226)
	{
		printf("send_file: La transmission du fichier a echoue\n");
		return -1;
	}
	
	printf("%.1f ko/s\n", ((float)total/1024)/(time(NULL)-tdeb));
	return 0;
}

/*
Analyse une reponse a une commande PASV :
Renvoie l'adresse IP et le port sur lequel
le client doit se connecter
*/
unsigned short parse_pasv_cmd(char *cmd, unsigned long *ip)
{
	int i;
	int pow=0;
	char *p;
	unsigned short port;
	port=0;
	*ip=0;

	if ((p=strchr(cmd, ')'))==NULL)
		return 0;
	*p=0x00;

	if ((cmd=strchr(cmd, '('))==NULL)
		return 0;
	cmd++;
	
	for (i=0; i<6; i++)
	{
		if ((p=strchr(cmd, ','))==NULL)
		{
			if (i<5)
				return 0;
		}
		else
			*p=0x00;
		
		if (i<4)
			*ip+=(atoi(cmd)<<pow);
		else
			port=(port<<8)+atoi(cmd);;
		cmd=p+1;
		pow+=8;
		if (i==3)
			pow=0;
	}

	return port;
}

/*
Fermeture de la connexion
*/
void ftp_disconnect(void)
{
	if (sock==-1)
		return;

	skprintf("QUIT\r\n");
	
	/*read_answer_code();*/

	printf("Disconnected\n");

	closesocket(sock);
	sock=-1;

	if (logfile)
		fclose(logfile);
}

/*
Initialisation du gestionnaire de sockets
*/
int init_sockets(void)
{
	WSADATA wdata;
	sock=-1;
	return WSAStartup(((int)1<<8)+1, &wdata);
}
