/*
Synchronisation de fichiers sur un compte FTP
*/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include "ftpsync.h"
#include "ftpclient.h"

/*
*/
int debug;
int proj_count;
struct projet projets[3];
struct projet *currproj;
char local_dir[MAX_PATH];

int main(int argc, char **argv)
{
	int i;
	char buffer[128];

	debug=0;
	
	for (i=1; i<argc; i++)
	{
		if (*argv[i]=='-')
		{
			switch(argv[i][1])
			{
				case 'd':
					debug++;
					break;
				default:
					break;
			}
		}
	}

	init_sockets();

	load_projects();
	printf("Loaded %d projects\n", proj_count);

	//user_cmd("debug");

	//user_cmd("load tuxfamily");
	//user_cmd("load local");
	
	//user_cmd("sync");

	do
	{
		if (currproj==NULL)
			printf("$ ");
		else
			printf("[%s]$ ", currproj->name);
		gets(buffer);
	}
	while (user_cmd(buffer));

	return 0;
}

/*
*/
int load_projects(void)
{
	FILE *conf;
	char buffer [256+1];
	char *tag, *val;
	int section=INTER_SECTION;

	if ((conf=fopen("ftpsync.conf", "rb"))==NULL)
	{
		printf("Le fichier de configuration est introuvable\n");
		return -1;
	}
	
	proj_count=0;
	currproj=NULL;

	memset(projets, 0, 3*sizeof(struct projet));

	while (fgets(buffer, 256, conf)!=NULL)
	{
		tag=buffer;
		if ((val=get_marker(&tag))!=NULL)
		{
			/*printf("tag=%s val=%s\n", tag, val);*/
			if (section==INTER_SECTION)
			{
				if (strcmp(tag, "projet")==0)
				{
					strcpy(projets[proj_count].name, val);
					section=PROJET;
				}
			}
			else if (section==PROJET)
			{
				if (strcmp(tag, "config")==0)
					section=CONFIG;
				else if (strcmp(tag, "/projet")==0)
				{
					proj_count++;
					section=INTER_SECTION;
				}
			}
			else if (section==CONFIG)
			{
				if (strcmp(tag, "host")==0)
					strcpy(projets[proj_count].host, val);
				else if (strcmp(tag, "user")==0)
					strcpy(projets[proj_count].user, val);
				else if (strcmp(tag, "pass")==0)
					strcpy(projets[proj_count].pass, val);
				else if (strcmp(tag, "lastsync")==0)
					sscanf(val, "%u:%u", &projets[proj_count].ft.dwHighDateTime, 
						&projets[proj_count].ft.dwLowDateTime);
				else if (strcmp(tag, "hostdir")==0)
					strcpy(projets[proj_count].host_dir, val);
				else if (strcmp(tag, "/config")==0)
					section=PROJET;
			}
		}
	}

	fclose(conf);
	return 0;
}

/*
*/
char *get_marker(char **marker)
{
	char *p, *q;

	if ((*marker=strchr(*marker, '<'))==NULL)
		return NULL;

	if ((p=strchr(*marker, '>'))==NULL)
		return NULL;

	(*marker)++;

	*p=0x00;

	if ((q=strchr(*marker, ' '))==NULL)
		return p;

	*q=0x00;
	return (q+1);
}

/*
*/
int user_cmd(char *buffer)
{
	char cmd[128];
	char *p;
	char *argv[5];
	int argc;

	strcpy(cmd, buffer);

	p=cmd;
	argc=0;
	argv[0]=cmd;
	do
	{
		if (argc>0)
		{
			*p=0x00;
			argv[argc]=p+1;
		}
		argc++;
	}
	while ((p=strchr(p+1, ' '))!=NULL);

	if (strcmp(cmd, "debug")==0)
	{
		debug^=1;
		printf("Debug : %s\n", debug?"On":"Off");
	}
	else if (strcmp(cmd, "load")==0)
	{
		if (argc==2)
			currproj=find_project(argv[1]);
	}
	else if (strcmp(cmd, "ls")==0)
		print_projects();
	else if (strcmp(cmd, "quit")==0)
		return 0;
	else if (strcmp(cmd, "sync")==0)
		sync_project(currproj);

	return 1;
}

/*
*/
struct projet *find_project(char *name)
{
	int i;

	for (i=0; i<proj_count; i++)
		if (strcmp(projets[i].name, name)==0)
			return (projets+i);

	printf("Le projet %s est introuvable\n", name);
	return NULL;
}

/*
*/
void print_projects(void)
{
	int i;

	for (i=0; i<proj_count; i++)
		printf("[%d] %s (%s)\n", i+1, projets[i].name, projets[i].host); 
}

/*
*/
int sync_project(struct projet *proj)
{
	FILE *conf;
	char buffer[256+1];
	char *tag, *val;
	int section=INTER_SECTION;
	int erreurs=0;
	HANDLE hfile;
	struct file_properties fprop;

	*local_dir=0x00;

	if (proj==NULL)
		return -1;

	if ((conf=fopen("ftpsync.conf", "rb"))==NULL)
		return -1;

	if (ftp_init(proj)==-1)
	{
		fclose(conf);
		return -1;
	}

	while (fgets(buffer, 256, conf)!=NULL)
	{
		tag=buffer;
		if ((val=get_marker(&tag))!=NULL)
		{
			if (section==INTER_SECTION)
			{
				if (strcmp(tag, "projet")==0)
					if (strcmp(val, proj->name)==0)
						section=PROJET;
			}
			else if (section==PROJET)
			{
				if (strcmp(tag, "/projet")==0)
					return -1;
				else if (strcmp(tag, "files")==0)
					section=FILES;
			}
			else if (section==FILES)
			{
				if (strcmp(tag, "/files")==0)
				{
					if (erreurs==0)
						update_lastsync(proj);
					fclose(conf);
					save_projects();
					ftp_disconnect();
					return 0;
				}
				if (strcmp(tag, "file")==0)
				{
					if (get_file_properties(val, &fprop))
					{
						hfile=sync_file(&fprop);
						if (hfile!=INVALID_HANDLE_VALUE)
						{
							if (send_file(hfile, &fprop)!=0)
								erreurs++;
							CloseHandle(hfile);
						}
					}
				}
				else if (strcmp(tag, "directory")==0)
				{
					get_file_properties(val, &fprop);
					ls_directory(&fprop);
				}
				else if (strcmp(tag, "setlocaldir")==0)
				{
					strcpy(local_dir, val);
				}
				else if (strcmp(tag, "/setlocaldir")==0)
				{
					*local_dir=0x00;
				}
			}
		}
	}

	fclose(conf);
	ftp_disconnect();
	return 0;
}

/*
*/
int get_file_properties(char *prop, struct file_properties *fprop)
{
	char *next;
	char *val;

	*fprop->local=0x00;
	*fprop->remote=0x00;
	fprop->chmod=0;
	fprop->recursive=0;

	do
	{
		next=strchr(prop, ' ');
		if (next)
			*next=0x00;

		if ((val=strchr(prop, '='))!=NULL)
		{
			*val=0x00;
			val++;
			if (strcmp(prop, "local")==0)
				sprintf(fprop->local, "%s%s", local_dir, val);
			else if (strcmp(prop, "serv")==0)
				strcpy(fprop->remote, val);
			else if (strcmp(prop, "chmod")==0)
				fprop->chmod=atoi(val);
		}
		if (strcmp(prop, "recursive")==0)
				fprop->recursive=1;
		prop=next+1;
	}
	while (next);

	return (*fprop->local!=0x00 && *fprop->remote!=0x00);
}

/*
*/
HANDLE sync_file(struct file_properties *fprop)
{
	HANDLE hfile;
	FILETIME lft;

	hfile=CreateFile(fprop->local, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);

	if (hfile==INVALID_HANDLE_VALUE)
	{
		printf("%s introuvable\n", fprop->local);
		return INVALID_HANDLE_VALUE;
	}

	GetFileTime(hfile, NULL, NULL, &lft);

	/*FileTimeToLocalFileTime(&ft, &lft);*/

	if (CompareFileTime(&lft, &currproj->ft)==-1)
	{
		CloseHandle(hfile);
		return INVALID_HANDLE_VALUE;
	}

	return hfile;
}

/*
*/
int ls_directory(struct file_properties *fprop)
{
	WIN32_FIND_DATA wfdata;
	HANDLE hlist, hfile;
	short local_len;
	short chdir_done = 0;

	local_len=strlen(fprop->local);

	if (fprop->remote[0]==0x00)
		chdir_done=2;
	
	strcat(fprop->local, "*");

	if ((hlist=FindFirstFile(fprop->local, &wfdata))==INVALID_HANDLE_VALUE)
	{
		printf("Erreur FindFirstFile\n");
		return -1;
	}

	do
	{
		/* Si ce n'est pas un dossier */
		if ((wfdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)==0)
		{
			strcpy(fprop->local+local_len, wfdata.cFileName);
			if ((hfile=sync_file(fprop))!=INVALID_HANDLE_VALUE)
			{
				if (!chdir_done)
				{
					chdir(fprop->remote);
					chdir_done=1;
				}
				strcpy(fprop->remote, wfdata.cFileName);
				send_file(hfile, fprop);
				CloseHandle(hfile);
			}
		}
		else
		{
			if (fprop->recursive && wfdata.cFileName[0]!='.')
			{
				sprintf(fprop->local+local_len, "%s\\", wfdata.cFileName);
				if (!chdir_done)
				{
					chdir(fprop->remote);
					chdir_done=1;
				}
				strcpy(fprop->remote, wfdata.cFileName);
				ls_directory(fprop);
			}
		}
	}
	while (FindNextFile(hlist, &wfdata)!=0);

	FindClose(hlist);

	fprop->local[local_len]=0x00;

	if (chdir_done==1)
		chdir("..");
	fprop->remote[0]=0x00;

	return 0;
}

/*
*/
void update_lastsync(struct projet *proj)
{
	SYSTEMTIME st;
	
	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &proj->ft);
}

/*
*/
int save_projects(void)
{
	FILE *conf, *dst;
	char buffer [256+1];
	char *tag, *val;
	int section=INTER_SECTION;
	struct projet *proj;

	if ((conf=fopen("ftpsync.conf", "rb"))==NULL)
		return -1;

	if ((dst=fopen("ftpsync.tmp", "wb"))==NULL)
		return -1;

	while (fgets(buffer, 256, conf)!=NULL)
	{
		tag=buffer;
		if ((val=get_marker(&tag))!=NULL)
		{
			if (section==INTER_SECTION)
			{
				if (strcmp(tag, "projet")==0)
				{
					fprintf(dst, "<projet %s>\r\n", val); 
					proj=find_project(val);
					section=PROJET;
					fprintf(dst, "<config>\r\n");
					fprintf(dst, "<host %s>\r\n", proj->host);
					fprintf(dst, "<user %s>\r\n", proj->user);
					fprintf(dst, "<pass %s>\r\n", proj->pass);
					fprintf(dst, "<lastsync %u:%u>\r\n", proj->ft.dwHighDateTime, proj->ft.dwLowDateTime);
					fprintf(dst, "<hostdir %s>\r\n", proj->host_dir);
					fprintf(dst, "</config>\r\n");
				}
			}
			else if (section==PROJET)
			{
				if (strcmp(tag, "config")==0)
					section=CONFIG;
				else
				{
					if (strcmp(tag, "/projet")==0)
						section=INTER_SECTION;

					if (*val!=0x00)
						fprintf(dst, "<%s %s>\r\n", tag, val);
					else
						fprintf(dst, "<%s>\r\n", tag);
				}
			}
			else if (section==CONFIG)
			{
				if (strcmp(tag, "/config")==0)
					section=PROJET;
			}
		}
		else
		{
			fprintf(dst, "%s", buffer);
		}
	}

	fclose(conf);
	fclose(dst);

	DeleteFile("ftpsync.conf");
	MoveFile("ftpsync.tmp", "ftpsync.conf");

	return 0;
}

/*
*/
void to_system_time(FILETIME *ft, char *buffer)
{
	SYSTEMTIME st;

	FileTimeToSystemTime(ft, &st);
	
	sprintf(buffer, "%d/%d/%d/%d %d:%d:%d", st.wDay, st.wMonth, st.wYear, st.wDayOfWeek,
		st.wHour, st.wMinute, st.wSecond);
}

/*
*/
void to_filetime(char *buffer, FILETIME *ft)
{
	SYSTEMTIME st;
	sscanf(buffer, "%d/%d/%d/%d %d:%d:%d", &st.wDay, &st.wMonth, &st.wYear,
		&st.wDayOfWeek, &st.wHour, &st.wMinute, &st.wSecond);

	printf("%d\n", st.wYear);

	SystemTimeToFileTime(&st, ft);
}
